extern "C" {
#include "io.h"
#include "talk.h"
}
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <setjmp.h>

const int height = 23, width = 10;

static void timer_stop();
static void timer_continue();

class cluster;

class field {
	int field_x, field_y;
	char map[width+6][height+3];
	WINDOW *window;
	friend void transfer(field *f1, field *f2, cluster *c2);
	char gift[width][4];
	int left_time;
	int gift_lines;
public:
	field(int fx, int fy, char *title);
	~field() { delwin(window); }
	frefresh() { wrefresh(window); }
	void fplot(int x, int y, char *s) { mvwaddstr(window, y, x*2+2, s); }
	void fset(int x, int y, int c) { map[x+3][y] = c; }
	int fget_map(int x, int y) { return map[x+3][y]; }
	int delayed_appearance();
	int scrollup();
	int scrolldown(int);
	void update();
};

field::field(int fx, int fy, char *title)
{
	int i, j;
    field_x = fx;
	field_y = fy;
	left_time = -1;
	window = newwin(height+1, width*2+14, fy, fx);
	scrollok(window, FALSE);
	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			fset(j, i, 0);
	for (i = 0; i < height+1; i++) {
		fplot(-1, i, "%%");
		fset(-1, i, 5);
		for (j = width; j < width+6; j++) {
			fplot(j, i, "%%");
		}
		fset(width, i, 5);
	}
	for (i = 0; i < width; i++) {
		fplot(i, height, "%%");
		fset(i, height, 5);
	}
	for (i = 0; i < 4; i++)
		for (j = width+1; j < width+5; j++)
			fplot(j, i, "  ");
	fplot(width+1, height-6, "Left  :J");
	fplot(width+1, height-5, "Right :L");
	fplot(width+1, height-4, "Rotate:K");
	fplot(width+1, height-3, "Drop  : ");
	fplot(width+1, height-2, "Quit  :Q");
	fplot((width+5)/2 - strlen(title)/4, height, title);
}

inline int field::scrollup()
{
	for (int i = 0; i < width; i++)
		if (fget_map(i, 0))
			return -1;
	for (i = 0; i < height-1; i++)
		for (int j = 0; j < width; j++)
			fset(j, i, fget_map(j, i+1));
	return 0;
}

inline int field::scrolldown(int l)
{
	for (; l > 1; l--)
		for (int i = 0; i < width; i++)
			fset(i, l, fget_map(i, l-1));
}

inline static void transfer(field *f1, field *f2, cluster *c2)
{
	if (f2->left_time < 0)
		f2->gift_lines = 0;
	for (int i = 0; i < height; i++) {
		int full = 1;
		for (int j = 0; j < width; j++)
			if (f1->fget_map(j, i) == 0) {
				full = 0;
				break;
			}
		if (full) {
			if (f2->left_time < 0) {
				for (j = 0; j < width; j++) {
					f2->gift[j][f2->gift_lines] = (f1->fget_map(j, i) == 1) ? 1 : 0;
				}
				f2->gift_lines++;
			}
			f1->scrolldown(i);
		}
	}
	if (f2->left_time < 0 && f2->gift_lines > 1)
		f2->left_time = 2;
	f1->update();
	f1->frefresh();
}

inline void field::update()
{
	for (int i = 0; i < height; i++)
		for (int j = 0; j < width; j++) {
			fset(j, i, !!fget_map(j, i));
			if (fget_map(j, i))
				fplot(j, i, "[]");
			else fplot(j, i, "  ");
		}
}

static class cluster *my_next;
static char queueorsocket();
static char sread();
static char swrite();
static int set_current(cluster *, cluster *, field *, int);
static int rand_block();
static void game_over(char);
static jmp_buf battle_buf;

class cluster {
	const static char pattern[7][4][4][4];
	int contact_time;
	int x, y;
	int shape;
	int direction;
	field *f;
public:
	cluster(field *fd, int cx, int cy, int cs, int d)
	    { contact_time = 0; x = cx; y = cy; shape = cs; direction = d; f = fd; }
	void cdraw();
	void cerase();
	void clocate(int cx, int cy) { x = cx; y = cy; }
	int canbeat(int cx, int cy, int d);
	int down();
	int drop(cluster *, field *, int);
	void left();
	void right();
	void rotate();
	void set_shape(int s) { shape = s; }
	void set_direction(int d) { direction = d; }
	int get_shape() { return shape; }
	int get_direction() { return direction; }
	int get_x() { return x; }
	int get_y() { return y; }
};

inline void cluster::cdraw()
{
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			if (pattern[shape][direction][i][j])
				f->fplot(x+j, y+i, "[]");
}

inline void cluster::cerase()
{
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			if (pattern[shape][direction][i][j])
				f->fplot(x+j, y+i, "  ");
}

inline int cluster::canbeat(int cx, int cy, int d)
{
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			if (pattern[shape][d][i][j] && f->fget_map(cx+j, cy+i))
				return 0;
	return 1;
}
	
inline void cluster::rotate()
{
	int d = (direction+1) % 4;

	if (canbeat(x, y, d)) {
		cerase();
		direction = d;
		cdraw();
		f->frefresh();
	}
}

inline void cluster::left()
{
	if (canbeat(x-1, y, direction)) {
		cerase();
		clocate(x-1, y);
		cdraw();
		f->frefresh();
	}
}

inline void cluster::right()
{
	if (canbeat(x+1, y, direction)) {
		cerase();
		clocate(x+1, y);
		cdraw();
		f->frefresh();
	}
}

inline int cluster::drop(cluster *next, field *other, int next_shape)
{
	cerase();
	for (int i = y; canbeat(x, i+1, direction); i++)
		;
	clocate(x, i);
	for (int j = 0; j < 4; j++)
		for (int k = 0; k < 4; k++)
			if (pattern[shape][direction][j][k])
				f->fset(x+k, y+j, 2);
	cdraw();
	f->frefresh();
	transfer(f, other, this);
	return set_current(this, next, f, next_shape);
}

extern int errno;
static int master;
static field *his_field, *my_field;
static cluster *my_current, *his_current, *my_next, *his_next;
static struct itimerval itimer_save;
static int nb;
static char buf[BUFSIZ];
static char *lqp;
static int ended;
static int his_absence, my_absence;

inline int field::delayed_appearance()
{
	if (left_time < 0) return 0;
	else if (left_time-- == 0) {
		for (int i = 0; i < gift_lines; i++) {
			if (scrollup()) { update(); frefresh(); return -1; }
			for (int j = 0; j < width; j++)
				fset(j, height-1, gift[j][i]);
		}
		update();
		return 1;
	}
	return 0;
}

inline int cluster::down()
{
	if (canbeat(x, y+1, direction)) {
		contact_time = 0;
		cerase();
		clocate(x, y+1);
		cdraw();
		f->frefresh();
		return 0;
	}
	else
		if (++contact_time > 2) {
			contact_time = 0;
			return 1;
		} else return 0;
}

static int prepare();
static void battle();
static void cleanup();
static void tick();
static void slave_tick();
static void timer_start();
static void timer_off();
static void socket_process(int);
static void key_process(int);

extern "C"
void bt(int master)
{
	::master = master;
	if (prepare() == 0)
		battle();

	cleanup();
}

inline static char queueorsocket()
{
	char c;

	if (nleft > 0) {
		nleft--;
		return *lqp++;
	}
	else {
		read(sockt, &c, 1);
		return c;
	}
}

static int prepare()
{
	enum error_code { good, version, screen_size, simultaneous };
	const char *error_message[] = {
		"",
		"This program is too old",
		"His screen is too small",
		"You and he typed #TETRIS simultaneously"
	};
	char c;
	const int my_version = 0x0022;
	int his_version = 0;
	char my_prob, his_prob;

	lqp = leftqueue;
	buf[0] = 0;
	buf[1] = 'b';
	buf[2] = 't';
	write(sockt, buf, 3);

	for (;;) {
		c = queueorsocket();
		if (c == '\0') {
			c = queueorsocket();
			if (c == 'b')
				break;
		}
	}
	if (queueorsocket() != 't') {	// his is 0.1
		const char notice[] = "\000\000qo\002Your hbtalk is version 0.1."
		                      " It has a serious bug.";
		message("His hbtalk's version is 0.1");
		write(sockt, notice, sizeof notice-1);
		for (int i = 4; i > 0; i--)
			queueorsocket();
		return -1;
	}

	buf[0] = my_version >> 8;
	buf[1] = my_version & 0xff;
	buf[2] = master;
	write(sockt, buf, 3);

	his_version = (unsigned char)queueorsocket();
	his_version = his_version << 8 | queueorsocket();

	my_prob = good;
	if (my_version > his_version) {
		my_prob = version;
		message("His hbtalk is older than version 0.22");
	}
	if (LINES < height+1 || COLS < (width+7)*2) {
		my_prob = screen_size;
		message("Screen is too small");
	}
	if (master == queueorsocket()) {
		my_prob = simultaneous;
		message("You and he typed #TETRIS simultaneously");
	}
	write(sockt, &my_prob, 1);
	if ((his_prob = queueorsocket()) != good)
		message(error_message[his_prob]);
	if (my_prob != good || his_prob != good)
		return -1;

	time_t t = time(NULL);
	srand(t);

	his_field = new field(0, 0, "Your Party's");
	my_field = new field(COLS/2, 0, "Yours");

	buf[0] = rand_block();
	write(sockt, buf, 1);
	my_current = new cluster(my_field, width/2-2, 0, buf[0], 0);
	my_current->cdraw();
	his_current = new cluster(his_field, width/2-2, 0, queueorsocket(), 0);
	his_current->cdraw();

	buf[0] = rand_block();
	write(sockt, buf, 1);
	my_next = new cluster(my_field, width+1, 0, buf[0], 0);
	my_next->cdraw();
	his_next = new cluster(his_field, width+1, 0, queueorsocket(), 0);
	his_next->cdraw();

	his_field->frefresh();
	my_field->frefresh();

	his_absence = my_absence = -1;

	return 0;
}

static void cleanup()
{
	delete his_field;
	delete my_field;

	touchwin(my_win.x_win);
	touchwin(his_win.x_win);
	touchwin(line_win);

	wrefresh(my_win.x_win);
	wrefresh(his_win.x_win);
	wrefresh(line_win);
}

static void battle()
{
	int sockt_mask = 1 << sockt;
	int read_template = sockt_mask | STDIN_MASK;
	int read_set;
	struct timeval wait;

	timer_start();

	ended = 0;
	setjmp(battle_buf);
	if (ended) goto bt_end;

	for (;;) {
		if (nleft > 0) {
			nleft--;
			socket_process(*lqp++);
			continue;
		}
		read_set = read_template;
		wait.tv_sec = A_LONG_TIME;
		wait.tv_usec = 0;
		nb = select(32, &read_set, 0, 0, &wait);
		if (nb <= 0) {

			/* We may be returning from an interupt handler */

			if (errno == EINTR) {
				read_set = read_template;
				continue;
			} else {
				/* panic, we don't know what happened */
				cleanup();
				p_error("Unexpected error from select");
				quit();
			}
		}

		if ( read_set & sockt_mask ) {

			/* There is data on sockt */
			nb = read(sockt, buf, 1);

			if (nb <= 0) {
				message("Connection closed. Exiting");
				quit();
			} else
				socket_process(buf[0]);
		}

		if ( read_set & STDIN_MASK ) {

			/* we can't make the tty non_blocking, because
		   curses's output routines would screw up */

			ioctl(0, FIONREAD, (struct sgttyb *) &nb);
			nb = read(0, buf, 1);
			key_process(buf[0]);
			/* We might lose data here because sockt is non-blocking */
		}
	}
bt_end:
	;
}

static void tick()
{
	timer_stop();

	int i_died = 0, he_died = 0;

	write(sockt, "T", 1);
	int d = my_field->delayed_appearance();
	switch(d) {
	case -1: i_died = 1; my_field->frefresh(); break;
	case  1: my_absence = 0; my_field->frefresh(); break;
	case  0:
		if (my_current->down()) {
			char c = rand_block();
			write(sockt, &c, 1);
			i_died = my_current->drop(my_next, his_field, c);
		}
		break;
	}
	if (my_absence != -1)
		if (++my_absence > 1) {
			my_current->cerase();
			char c = rand_block();
			write(sockt, &c, 1);
			i_died = set_current(my_current, my_next, my_field, c) == 2;
			my_field->frefresh();
			my_absence = -1;
		}

	switch(d = his_field->delayed_appearance()) {
	case -1: he_died = 1; his_field->frefresh(); break;
	case  1: his_absence = 0; his_field->frefresh(); break;
	case  0:
		if (his_current->down()) {
			char c = rand_block();
			write(sockt, &c, 1);
			he_died = his_current->drop(his_next, my_field, c);
		}
		break;
	}
	if (his_absence != -1)
		if (++his_absence > 1) {
			his_current->cerase();
			char c = rand_block();
			write(sockt, &c, 1);
			he_died = set_current(his_current, his_next, his_field, c) == 1;
			his_field->frefresh();
			his_absence = -1;
		}

	if (i_died && he_died) game_over(3);
	else if (i_died) game_over(2);
	else if (he_died) game_over(1);

	timer_continue();
}

inline static void slave_tick()
{
	int i_died = 0, he_died = 0;

	int d = his_field->delayed_appearance();
	switch(d) {
	case -1: he_died = 1; his_field->frefresh(); break;
	case  1: his_absence = 0; his_field->frefresh(); break;
	case  0:
		if (his_current->down())
			he_died = his_current->drop(his_next, my_field, queueorsocket());
		break;
	}
	if (his_absence != -1)
		if (++his_absence > 1) {
			his_current->cerase();
			char c = queueorsocket();
			he_died = set_current(his_current, his_next, his_field, c) == 1;
			his_field->frefresh();
			his_absence = -1;
		}

	d = my_field->delayed_appearance();
	switch(d) {
	case -1: i_died = 1; my_field->frefresh(); break;
	case  1: my_absence = 0; my_field->frefresh(); break;
	case  0:
		if (my_current->down())
			i_died = my_current->drop(my_next, his_field, queueorsocket());
		break;
	}
	if (my_absence != -1)
		if (++my_absence > 1) {
			my_current->cerase();
			char c = queueorsocket();
			i_died = set_current(my_current, my_next, my_field, c) == 2;
			my_field->frefresh();
			my_absence = -1;
		}

	if (i_died && he_died) game_over(3);
	else if (i_died) game_over(2);
	else if (he_died) game_over(1);
}

inline static void key_process(int c)
{
	timer_stop();

	switch (c) {
	case 'q': case 'Q':
		if (master) { write(sockt, "X", 1); game_over(0); }
		else write(sockt, "x", 1); 
		break;
	case 'k': case 'K':
		if (master) {
			if (my_absence == -1) { write(sockt, "O", 1); my_current->rotate(); }
		}
		else if (my_absence == -1) write(sockt, "o", 1);
		break;
	case 'j': case 'J':
		if (master) {
			if (my_absence == -1) { write(sockt, "L", 1); my_current->left(); }
		}
		else if (my_absence == -1) write(sockt, "l", 1);
		break;
	case 'l': case 'L':
		if (master) {
			if (my_absence == -1) { write(sockt, "R", 1); my_current->right(); }
		}
		else if (my_absence == -1) write(sockt, "r", 1);
		break;
	case ' ':
		if (master) {
			if (my_absence == -1) {
				char c = rand_block();
				write(sockt, "D", 1);
				write(sockt, &c, 1);
				if (my_current->drop(my_next, his_field, c))
					game_over(2);
			}
		}
		else if (my_absence == -1) write(sockt, "d", 1);
		break;
	}

	timer_continue();
}

inline void socket_process(int c)
{
	timer_stop();

	switch (c) {
	case 'T': slave_tick(); break;
	case 'X': game_over(0);
	case 'x': write(sockt, "Y", 1); game_over(0);
	case 'Y': game_over(0);
	case 'O': his_current->rotate(); break;
	case 'o': write(sockt, "P", 1); his_current->rotate(); break;
	case 'P': my_current->rotate(); break;
	case 'L': his_current->left(); break;
	case 'l': write(sockt, "M", 1); his_current->left(); break;
	case 'M': my_current->left(); break;
	case 'R': his_current->right(); break;
	case 'r': write(sockt, "S", 1); his_current->right(); break;
	case 'S': my_current->right(); break;
	case 'D': 
		if (his_current->drop(his_next, my_field, queueorsocket()))
			game_over(1);
		break;
	case 'd':
		char c = rand_block();
		write(sockt, "E", 1);
		write(sockt, &c, 1);
		if (his_current->drop(his_next, my_field, c)) game_over(1);
		break;
	case 'E':
		if (my_current->drop(my_next, his_field, queueorsocket())) game_over(2);
		break;
	}

	timer_continue();
}

inline int set_current(cluster *current, cluster *next, field *f, int s)
{
	current->set_shape(next->get_shape());
	current->set_direction(next->get_direction());
	current->clocate(width/2-2, 0);
	current->cdraw();
	next->cerase();
	next->set_shape(s);
	next->set_direction(0);
	next->cdraw();
	f->frefresh();
	if (!current->canbeat(current->get_x(), current->get_y(), current->get_direction()))
		return (current == my_current) ? 2 : 1;	// 2 if i die 1 if he die
	else return 0;
}

static void game_over(char result)
{
	timer_off();

	write(sockt, "V", 1);
	write(sockt, &result, 1);
	while (queueorsocket() != 'V')
		;
	int his_result = queueorsocket();
	his_field->fplot(width/2 - 5, height/2-1, "                    ");
	his_field->fplot(width/2 - 5, height/2+1, "                    ");
	my_field->fplot(width/2 - 5, height/2-1, "                    ");
	my_field->fplot(width/2 - 5, height/2+1, "                    ");
	if (his_result == 0 || result == 0) {
		his_field->fplot(width/2 - 5, height/2, "        QUIT        ");
		my_field->fplot(width/2 - 5, height/2, "        QUIT        ");
	}
	else if (his_result == 3 || result == 3 || his_result == result) {
		his_field->fplot(width/2 - 5, height/2, "         TIE        ");
		my_field->fplot(width/2 - 5, height/2, "        TIE         ");
	}
	else if (result == 2) {
		his_field->fplot(width/2 - 5, height/2, "         WON        ");
		my_field->fplot(width/2 - 5, height/2, "        LOST        ");
	}
	else {
		his_field->fplot(width/2 - 5, height/2, "        LOST        ");
		my_field->fplot(width/2 - 5, height/2, "        WON         ");
	}
	his_field->frefresh();
	my_field->frefresh();

	sleep(3);
	ended = 1;
	longjmp(battle_buf, 0);
}

inline static int rand_block()
{
	return rand() % 7;
}

static void timer_start()
{
	if (!master) return;

	struct itimerval itimer;

	itimer.it_value.tv_sec = 3;
	itimer.it_value.tv_usec = 0;
	itimer.it_interval.tv_sec = 1;
	itimer.it_interval.tv_usec = 0;
	setitimer(ITIMER_REAL, &itimer, (struct itimerval *)0);
	signal(SIGALRM, tick);
}

inline static void timer_stop()
{
	if (!master) return;

	struct itimerval itimer;

	timerclear(&itimer.it_value);
	timerclear(&itimer.it_interval);
	setitimer(ITIMER_REAL, &itimer, &itimer_save);
	if (itimer_save.it_value.tv_sec == 0 && itimer_save.it_value.tv_usec == 0)
		itimer_save.it_value.tv_usec = 1;
}

inline static void timer_continue()
{
	if (!master) return;

	setitimer(ITIMER_REAL, &itimer_save, (struct itimerval *)0);
}

static void timer_off()
{
	struct itimerval itimer;

	timerclear(&itimer.it_value);
	timerclear(&itimer.it_interval);
	setitimer(ITIMER_REAL, &itimer, (struct timerval *)0);
}

const char cluster::pattern[7][4][4][4] = {
	// 0
	0,0,0,0,
	1,1,1,1,
	0,0,0,0,
	0,0,0,0,

	0,0,1,0,
	0,0,1,0,
	0,0,1,0,
	0,0,1,0,

	0,0,0,0,
	1,1,1,1,
	0,0,0,0,
	0,0,0,0,

	0,0,1,0,
	0,0,1,0,
	0,0,1,0,
	0,0,1,0,
	// 1
	0,0,0,0,
	0,1,1,0,
	0,0,1,1,
	0,0,0,0,

	0,0,1,0,
	0,1,1,0,
	0,1,0,0,
	0,0,0,0,

	0,0,0,0,
	0,1,1,0,
	0,0,1,1,
	0,0,0,0,

	0,0,1,0,
	0,1,1,0,
	0,1,0,0,
	0,0,0,0,

	// 2
	0,0,0,0,
	0,1,1,0,
	1,1,0,0,
	0,0,0,0,

	0,1,0,0,
	0,1,1,0,
	0,0,1,0,
	0,0,0,0,

	0,0,0,0,
	0,1,1,0,
	1,1,0,0,
	0,0,0,0,

	0,1,0,0,
	0,1,1,0,
	0,0,1,0,
	0,0,0,0,

	// 3
	0,0,0,0,
	0,1,1,1,
	0,0,0,1,
	0,0,0,0,

	0,0,1,1,
	0,0,1,0,
	0,0,1,0,
	0,0,0,0,

	0,1,0,0,
	0,1,1,1,
	0,0,0,0,
	0,0,0,0,

	0,0,1,0,
	0,0,1,0,
	0,1,1,0,
	0,0,0,0,

	// 4
	0,0,0,0,
	0,1,1,1,
	0,1,0,0,
	0,0,0,0,

	0,0,1,0,
	0,0,1,0,
	0,0,1,1,
	0,0,0,0,

	0,0,0,1,
	0,1,1,1,
	0,0,0,0,
	0,0,0,0,

	0,1,1,0,
	0,0,1,0,
	0,0,1,0,
	0,0,0,0,

	// 5
	0,0,1,0,
	0,1,1,1,
	0,0,0,0,
	0,0,0,0,

	0,0,1,0,
	0,1,1,0,
	0,0,1,0,
	0,0,0,0,

	0,0,0,0,
	0,1,1,1,
	0,0,1,0,
	0,0,0,0,

	0,0,1,0,
	0,0,1,1,
	0,0,1,0,
	0,0,0,0,
	// 6
	0,1,1,0,
	0,1,1,0,
	0,0,0,0,
	0,0,0,0,

	0,1,1,0,
	0,1,1,0,
	0,0,0,0,
	0,0,0,0,

	0,1,1,0,
	0,1,1,0,
	0,0,0,0,
	0,0,0,0,

	0,1,1,0,
	0,1,1,0,
	0,0,0,0,
	0,0,0,0
};

